'''
This file provides all the functions and classes to handle painting
graphics and moving things: qcanvasitem generation, minimap generation,
moves, ...

Author: Philippe Fremy
License: Gnu GPL (see fname LICENSE)
'''
from typing import Optional, Tuple, List, Dict
import math, pathlib

from PyQt5.QtWidgets import QSizePolicy, QFrame, QGraphicsScene, QGraphicsView, QWidget, QGraphicsPixmapItem
from PyQt5.QtGui import QImage, QBitmap, QPixmap, QPainter, QBrush, QColor, QMouseEvent
from PyQt5.QtCore import QTimer, pyqtSignal, QSize, Qt

from .kl_enum import (TILE_FILE_NAME, TRANSP_COLOR, MINI_TILE_SIZE, Piece,
                      Tile, TILE_SIZE, PIX_SIZE, piece_tile, is_piece, mini_tile_colors)
from .kl_map import KLMap


def cos(a: Tuple[int, int], b: Tuple[int, int]) -> float:
    '''computes cos(angle(a,b)) = scalar(a,b) / ( abs(a) * abs(b) )'''
    scal = a[0] * b[0] + a[1] * b[1]
    na = math.sqrt(a[0] ** 2 + a[1] ** 2)
    nb = math.sqrt(b[0] ** 2 + b[1] ** 2)
    return float(scal) / (na * nb)


class KLBoard(QGraphicsView):

    sig_move = pyqtSignal(tuple)

    def __init__(self, parent: QWidget):
        QGraphicsView.__init__(self, parent)
        self.klmap = KLMap()
        self.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding))
        self.setLineWidth(2)
        self.setFrameShape(QFrame.NoFrame)
        self.scene_item_dict = {}  # type: Dict[str, List[QGraphicsPixmapItem]]
        self.m_scene = QGraphicsScene(self)

        self.is_draging = False
        self.is_clicking = False
        self.drag_pid = ''
        self.drag_x = -1
        self.drag_y = -1
        self.drag_timer = QTimer(self)
        self.drag_timer.setSingleShot(True)
        self.drag_timer.timeout.connect(self.set_draging)

    pix_tiles = None  # type: Optional[QPixmap]
    tiles_mask = None  # type: Optional[QBitmap]

    def load_tiles(self) -> None:
        """Load the tiles representation from the default filename and put the result and the bitmap mask
        in KLBoard.tiles_mask and KLBoard.pix_tiles"""
        img_tile = QImage(str(pathlib.Path(__file__).parent / TILE_FILE_NAME))

        w = img_tile.width()
        h = img_tile.height()
        assert w != 0 and h != 0, "%s gives a null image" % TILE_FILE_NAME

        # Build mask
        img_mask = QImage(w, h, QImage.Format_Mono)
        for x in range(w):
            for y in range(h):
                if img_tile.pixel(x, y) == TRANSP_COLOR:
                    img_mask.setPixel(x, y, 0)
                else:
                    img_mask.setPixel(x, y, 1)

        KLBoard.tiles_mask = QBitmap()
        if not KLBoard.tiles_mask.convertFromImage(img_mask, Qt.MonoOnly | Qt.ThresholdDither
                                                             | Qt.AvoidDither):   # type: ignore
            raise Exception("Could not convert to Bitmap !")

        KLBoard.pix_tiles = QPixmap()
        if not KLBoard.pix_tiles.convertFromImage(img_tile):
            raise Exception("Could not convert to Pixmap !")

    def del_s_wall(self, x: int, y: int) -> None:
        '''Called when a s_wall is hit by the heart piece, to hide it'''
        for cs in self.scene_item_dict[Piece.s_wall]:
            if cs.x() == x * TILE_SIZE and cs.y() == y * TILE_SIZE:
                cs.setVisible(False)

    def generate_mini_map(self, m: KLMap) -> QPixmap:
        '''Generate a QPixmap representing the map (to be used as an icon)'''
        pm = QPixmap(MINI_TILE_SIZE * (m.w + 2) - 1, MINI_TILE_SIZE * (m.h + 2) - 1)
        pm.fill(mini_tile_colors[Tile.space])
        p = QPainter(pm)

        # draw border
        brush = QBrush(mini_tile_colors[piece_tile[Piece.wall]])
        p.fillRect(0, 0, MINI_TILE_SIZE * (m.w + 2), MINI_TILE_SIZE - 1, brush)
        p.fillRect(0, 0, MINI_TILE_SIZE - 1, MINI_TILE_SIZE * (m.h + 2), brush)
        p.fillRect(0, MINI_TILE_SIZE * (m.h + 1), MINI_TILE_SIZE * (m.w + 2), MINI_TILE_SIZE - 1, brush)
        p.fillRect(MINI_TILE_SIZE * (m.w + 1), 0, MINI_TILE_SIZE - 1, MINI_TILE_SIZE * (m.h + 2), brush)

        # draw content
        p.translate(MINI_TILE_SIZE, MINI_TILE_SIZE)
        for y in range(m.h):
            for x in range(m.w):
                pid = m.pid(x, y)
                if is_piece(pid):
                    color = QColor(mini_tile_colors[piece_tile[Piece.pieces[0]]])
                else:
                    color = QColor(mini_tile_colors[piece_tile[pid]])
                p.fillRect(
                    x * MINI_TILE_SIZE, y * MINI_TILE_SIZE, MINI_TILE_SIZE - 1, MINI_TILE_SIZE - 1, QBrush(color)
                )

                rpid = m.pid(x + 1, y)
                rdpid = m.pid(x + 1, y + 1)
                dpid = m.pid(x, y + 1)

                p.setPen(color)
                if rpid == pid:
                    p.drawLine(
                        (x + 1) * MINI_TILE_SIZE - 1,
                        y * MINI_TILE_SIZE,
                        (x + 1) * MINI_TILE_SIZE - 1,
                        (y + 1) * MINI_TILE_SIZE - 2,
                    )
                if dpid == pid:
                    p.drawLine(
                        x * MINI_TILE_SIZE,
                        (y + 1) * MINI_TILE_SIZE - 1,
                        (x + 1) * MINI_TILE_SIZE - 2,
                        (y + 1) * MINI_TILE_SIZE - 1,
                    )
                if rdpid == pid:
                    p.drawPoint((x + 1) * MINI_TILE_SIZE - 1, (y + 1) * MINI_TILE_SIZE - 1)

        p.end()
        return pm

    def generate_scene(self, m: KLMap) -> None:
        '''Generate a QGraphicScene object from a map and fills up the scene_item_dict'''
        new_scene = QGraphicsScene(0, 0, m.w * TILE_SIZE, m.h * TILE_SIZE)
        self.setScene(new_scene)
        del self.scene_item_dict
        if self.m_scene:
            del self.m_scene
        self.m_scene = new_scene
        self.scene_item_dict = {}

        #  set the background color <=> fill
        self.m_scene.setBackgroundBrush(Qt.black)

        pix = QPixmap(TILE_SIZE, TILE_SIZE)
        mask = QBitmap(TILE_SIZE, TILE_SIZE)

        for x in range(m.w):
            for y in range(m.h):
                pid = m.pid(x, y)
                if pid == Piece.space:
                    continue

                # We pass the pixmap and bitmap object to avoid recreating them
                # on each function call
                self.set_tile_pix(pix, mask, x, y)
                item = QGraphicsPixmapItem(pix)
                item.setX(x * TILE_SIZE)
                item.setY(y * TILE_SIZE)
                self.m_scene.addItem(item)

                # moving tiles are above walls, special walls
                # and goals
                if pid == Piece.goal or pid == Piece.wall or pid == Piece.s_wall:
                    item.setZValue(1)
                else:
                    item.setZValue(10)

                if pid not in self.scene_item_dict:
                    self.scene_item_dict[pid] = []
                self.scene_item_dict[pid].append(item)

                item.show()

    def set_map(self, m: KLMap) -> None:
        m.sig_del_s_wall.connect(self.del_s_wall)
        self.klmap = m
        # note: the size of the new map is picked up by the sizeHint()
        self.generate_scene(m)
        self.update()

    def sizeHint(self) -> QSize:
        return QSize(self.klmap.w * TILE_SIZE, self.klmap.h * TILE_SIZE)

    def set_tile_pix(self, pix: QPixmap, mask: QBitmap, x: int, y: int) -> None:
        """Create the pixmap for the part of the piece located at x,y .
        The pixmap is created inside the pix argument"""
        pid = self.klmap.pid(x, y)
        piece_type = pid
        if is_piece(piece_type):
            piece_type = Piece.pieces[0]

        offset_list = [(0, 0), (PIX_SIZE, 0), (PIX_SIZE, PIX_SIZE), (0, PIX_SIZE)]

        # Value assiociated with each neighbour
        #
        #   2 | 1  1 | 2
        #   -----  -----
        #   4 | XXXX | 4
        #       XXXX
        #   4 | XXXX | 4
        #   -----  -----
        #   2 | 1  1 | 2
        #
        # (dx, dy, val)
        neighbour_val_list = [
            [(0, -1, 1), (-1, -1, 2), (-1, 0, 4)],
            [(0, -1, 1), (1, -1, 2), (1, 0, 4)],
            [(0, 1, 1), (1, 1, 2), (1, 0, 4)],
            [(0, 1, 1), (-1, 1, 2), (-1, 0, 4)],
        ]

        # which pixmap to choose according to the value of the
        #   neighbours
        which_quarter = {
            0: Tile.corner,
            1: Tile.ver_edge,
            2: Tile.corner,
            3: Tile.ver_edge,
            4: Tile.hor_edge,
            5: Tile.miss_corner,
            6: Tile.hor_edge,
            7: Tile.inner,
        }

        p = QPainter(pix)
        pm = QPainter(mask)
        for i in range(4):
            offset = offset_list[i]
            neighbour_val = neighbour_val_list[i]

            pix_nb = 0
            for neigh in neighbour_val:
                if pid == self.klmap.pid(x + neigh[0], y + neigh[1]):
                    pix_nb = pix_nb + neigh[2]

            quarter_nb = which_quarter[pix_nb]

            assert KLBoard.pix_tiles  # to help mypy making sure that tiles_mask is not None
            p.drawPixmap(
                offset[0],
                offset[1],
                KLBoard.pix_tiles,
                quarter_nb * TILE_SIZE + offset[0],
                piece_tile[piece_type] * TILE_SIZE + offset[1],
                PIX_SIZE,
                PIX_SIZE,
            )

            assert KLBoard.tiles_mask  # to help mypy making sure that tiles_mask is not None
            pm.drawPixmap(
                offset[0],
                offset[1],
                KLBoard.tiles_mask,
                quarter_nb * TILE_SIZE + offset[0],
                piece_tile[piece_type] * TILE_SIZE + offset[1],
                PIX_SIZE,
                PIX_SIZE,
            )

        p.end()
        pm.end()
        pix.setMask(mask)

    def move_piece(self, pid: str, delta: Tuple[int, int]) -> None:
        dx, dy = delta
        sz = len(self.scene_item_dict[pid])

        # For big tiles, we need to move big step by big step to keep smooth
        # display. For small tiles, we can move pixel by pixel.
        if sz < 3:
            incr = 1
        elif sz < 8:
            incr = 2
        elif sz < 16:
            incr = 4
        elif sz < 32:
            incr = 8
        else:
            incr = 16

        for _ in range(TILE_SIZE // incr):
            for cs in self.scene_item_dict[pid]:
                cs.moveBy(float(dx * incr), float(dy * incr))
            self.m_scene.update()

        incr = TILE_SIZE % incr
        if incr:
            for cs in self.scene_item_dict[pid]:
                cs.moveBy(float(dx * incr), float(dy * incr))
            self.m_scene.update()

    def mousePressEvent(self, e: QMouseEvent) -> None:
        pid = self.klmap.pid(e.x() // TILE_SIZE, e.y() // TILE_SIZE)
        if len(self.klmap.possibleMove(pid)) == 0:
            return

        self.is_clicking = True
        self.is_draging = False
        self.drag_pid = pid
        self.drag_x = e.x() // TILE_SIZE
        self.drag_y = e.y() // TILE_SIZE
        self.drag_timer.start(500)

    def set_draging(self) -> None:
        assert self.is_clicking, "timer expired while not clicking!!!"
        self.is_draging = True
        self.setCursor(Qt.ClosedHandCursor)

    def mouseReleaseEvent(self, _e: QMouseEvent) -> None:
        self.setCursor(Qt.ArrowCursor)
        if not self.is_clicking:
            return

        pm = self.klmap.possibleMove(self.drag_pid)
        if len(pm) == 0:
            self.drag_timer.stop()
            self.is_clicking = False
            return

        # tile was clicked
        if not self.is_draging:
            self.drag_timer.stop()
            if len(pm) == 1:
                self.sig_move.emit((self.drag_pid, pm[0]))
            self.is_clicking = False
            return

        # tile was dragged
        self.drag_timer.stop()
        self.is_clicking = False
        self.is_draging = False

    def mouseMoveEvent(self, e: QMouseEvent) -> None:
        if not self.is_clicking:
            return

        dx = e.x() // TILE_SIZE - self.drag_x
        dy = e.y() // TILE_SIZE - self.drag_y
        d = dx, dy
        if d == (0, 0):
            return

        pm = self.klmap.possibleMove(self.drag_pid)

        # computes the cos of the angle between d and the possible move
        l0 = map(lambda v, m: (cos(v, m), m), [d] * len(pm), pm)

        # keeps only moves in almost the same direction (cos > 0)
        l1 = filter(lambda x: x[0] > 0, l0)
        l2 = sorted(l1, reverse=True)
        rpm = list(map(lambda x: x[1], l2))

        if len(rpm) > 0:
            self.sig_move.emit((self.drag_pid, rpm[0]))
            self.is_draging = True
            self.drag_x = self.drag_x + rpm[0][0]
            self.drag_y = self.drag_y + rpm[0][1]



'''
KLMap is the internal structure used to store the info on the boards.

Author: Philippe Fremy
License: Gnu GPL (see fname LICENSE)
'''
from typing import Tuple, Dict, List, Optional
import copy
from functools import reduce

from PyQt5.QtCore import QObject, pyqtSignal

from .kl_enum import *


class KLMap(QObject):

    sig_del_s_wall = pyqtSignal(int, int)

    def __init__(self) -> None:
        QObject.__init__(self)
        self.xymap = []  # type: List[ List[str] ]
        self.goal = []  # type: List[ Tuple[int,int] ]
        self.s_wall = []  # type: List[ Tuple[int, int] ]

        self.name = ""
        self.w = -1
        self.h = -1
        self.menu_name = ""
        self.pid_size = 1
        self.orig_xymap = []  # type: List[ List[str] ]

    def __repr__(self) -> str:
        return self.to_string(self.xymap)

    def to_string(self, m: List[List[str]]) -> str:
        r = "xymap %s : %d x %d\n" % (self.name, self.w, self.h)
        for row in m:
            r = r + Piece.border
            for i in row:
                r = r + (' ' * self.pid_size + i)[-self.pid_size :]
            r = r + Piece.border + '\n'
        return r

    def reset(self) -> None:
        '''Reset the current xymap to its original value'''
        self.xymap = copy.deepcopy(self.orig_xymap)

    def pid(self, x: int, y: int) -> str:
        '''Return the piece id located at x,y'''
        if x < 0 or y < 0:
            return Piece.none
        if x >= self.w or y >= self.h:
            return Piece.none
        return self.xymap[y][x]

    def canMove(self, pid: str, dx: int, dy: int) -> bool:
        if abs(dx) + abs(dy) != 1:
            raise ValueError("test of illegal move")

        if not (pid == Piece.heart or is_piece(pid)):
            return False

        for x in range(self.w):
            for y in range(self.h):
                if self.pid(x, y) != pid:
                    continue
                move_id = self.pid(x + dx, y + dy)

                # moving outside the board is forbidden
                if move_id == Piece.none:
                    return False

                # moving through the door is ok for the heart piece
                if move_id == Piece.s_wall and pid == Piece.heart:
                    continue

                # moving onto oneself is ok
                # moving into space is ok
                # moving into final goal is ok
                if move_id == pid or move_id == Piece.space or move_id == Piece.goal:
                    continue

                # all other cases (moving onto other pieces) are forbidden
                return False

        return True

    def move_piece(self, pid: str, deltaxy: Optional[Tuple[int, int]] = None) -> None:
        """Move the piece pid by delta_x, delta_y wrapped in deltaxy

        If deltaxy is None, check of all possible moves and only move if one actual move is possible, else raises an exception.
        """
        if not (pid == Piece.heart or is_piece(pid)):
            raise Exception(str(pid) + " is not a piece, can not move it")

        if deltaxy is None:
            pm = self.possibleMove(pid)
            if len(pm) == 0:
                raise Exception("No possible move")
            if len(pm) > 1:
                raise Exception("Too many possible move")
            (dx, dy) = pm[0]
        else:
            dx, dy = deltaxy

        move_xymap = copy.deepcopy(self.xymap)
        pid_pos = []
        for x in range(self.w):
            for y in range(self.h):
                if self.xymap[y][x] == pid:
                    move_xymap[y][x] = Piece.space
                    pid_pos.append((x, y))

        for p in pid_pos:
            move_xymap[p[1] + dy][p[0] + dx] = pid
        for p in self.goal:
            if move_xymap[p[1]][p[0]] == Piece.space:
                move_xymap[p[1]][p[0]] = Piece.goal

        to_be_del_idx = []
        for i in range(len(self.s_wall)):
            p = self.s_wall[i]
            if move_xymap[p[1]][p[0]] == Piece.space:
                move_xymap[p[1]][p[0]] = Piece.s_wall
            elif move_xymap[p[1]][p[0]] == Piece.heart:
                # no longer a s_wall
                to_be_del_idx.append(i)

        to_be_del_idx.reverse()
        for i in to_be_del_idx:
            self.sig_del_s_wall.emit(*self.s_wall[i])
            del self.s_wall[i]

        self.xymap = move_xymap

    def possibleMove(self, pid: str) -> List[Tuple[int, int]]:
        '''Return the list of allowed moves for a given piece'''
        m = []  # type: List[ Tuple[int,int] ]
        for move in ((1, 0), (0, 1), (-1, 0), (0, -1)):
            if self.canMove(pid, move[0], move[1]):
                m.append(move)
        return m

    def whole_tile(self, x: int, y: int) -> List[Tuple[int, int]]:
        '''returns all the tiles belonging to the piece in x,y'''
        pid = self.pid(x, y)
        to_be_tested = [(x, y)]
        tested = {}  # type: Dict[Tuple[int, int], int]
        while len(to_be_tested):
            x, y = to_be_tested[-1]
            tested[(x, y)] = 0
            to_be_tested[-1:] = []

            # find adjacent tile of the same piece
            adj = []  # type: List[ Tuple[int, int]]
            for delta in [(0, 1), (1, 0), (-1, 0), (0, -1)]:
                if self.pid(x + delta[0], y + delta[1]) == pid:
                    adj.append((x + delta[0], y + delta[1]))

            for delta in adj:
                # Add d to to_be_tested list if d is not in tested
                # and if d is not in to_be_tested
                if delta not in tested:
                    if not delta in to_be_tested:
                        to_be_tested.append(delta)
        return list(tested.keys())

    def remove_doublon(self) -> None:
        def generate_new_pid(frompid: str) -> str:
            '''Generates a new identifier string which is not already defined in pid_dict'''
            while 1:
                if frompid in pid_dict:
                    frompid = frompid + pid[0]
                else:
                    return frompid

        self.pid_size = 1
        pid_dict = {}  # type: Dict[str, int]
        new_xymap = copy.deepcopy(self.xymap)

        for y in range(self.h):
            for x in range(self.w):
                pid = self.pid(x, y)
                if not is_piece(pid):
                    continue

                if pid in pid_dict:
                    new_pid = generate_new_pid(pid)
                    self.pid_size = max(self.pid_size, len(new_pid))
                else:
                    # pid_dict does not already contain pid
                    new_pid = pid

                pid_dict[new_pid] = 0
                l = self.whole_tile(x, y)
                for t in l:
                    tx, ty = t
                    new_xymap[ty][tx] = new_pid
                    # mark the piece as a wall to avoid reprocessing it
                    self.xymap[ty][tx] = Piece.s_wall

        self.xymap = new_xymap

    def check(self, line_no: int) -> None:
        """Check the general consistency of the current map.

        Raises an exception in case of inconsistency.

        Normally, all maps are correctly defined, exceptions should only be raised during
        development of new maps.
        """
        try:
            if len(self.name) == 0:
                raise ValueError("No name for level")

            if len(self.xymap) != self.h:
                raise AssertionError(
                    "xymap '%s' (line %d) has an inconsistent height %d." % (self.name, line_no, len(self.xymap))
                )
            for row in self.xymap:
                if len(row) != self.w:
                    raise AssertionError(
                        "xymap '%s' (line %d) has a line with inconsistent width %d.\nKLMap was supposed to be of width %d"
                        % (self.name, line_no, len(row), self.w)
                    )

            l = reduce(lambda x, y: x + y, self.xymap)
            s = reduce(lambda x, y: x + y, l, '')
            if -1 == s.find(Piece.heart) and self.name != "Icon":
                raise ValueError("xymap %s has no heart piece" % self.name)

            if -1 == s.find(Piece.goal) and self.name != "Icon":
                raise ValueError("xymap %s has no goal" % self.name)
        except (ValueError, AssertionError):
            print(self)
            raise

    def is_game_won(self) -> bool:
        '''Return True if the game is won'''
        for p in self.goal:
            if self.pid(p[0], p[1]) != Piece.heart:
                return False
        return True

    def loadline(self, line_nb: int, name: str) -> None:
        '''Fill the current xymap with one more line of data'''
        self.h = len(self.xymap)
        self.w = len(self.xymap[0])
        self.name = name
        self.check(line_nb)
        self.remove_doublon()

        self.goal = []
        self.s_wall = []
        for y in range(self.h):
            for x in range(self.w):
                if self.xymap[y][x] == Piece.goal:
                    self.goal.append((x, y))
                if self.xymap[y][x] == Piece.s_wall:
                    self.s_wall.append((x, y))

        self.orig_xymap = copy.deepcopy(self.xymap)


def load_maps(fname: str) -> Dict[int, KLMap]:
    '''Parses the file name to build a dictionnary of map id to KLMap objects'''
    f = open(fname)
    level_by_id = {}  # type: Dict[int, KLMap]
    intro = False
    klmap = KLMap()
    filling_map = False
    map_nb = -1
    line_nb = 0
    map_line_nb = 0
    map_name = ''

    for l in f.readlines():
        line_nb = line_nb + 1
        l = l[:-1].strip()
        if not len(l):
            continue

        if l[0] == Piece.border and (l.rfind(Piece.border) != l.find(Piece.border)):
            if not intro:
                continue
            filling_map = True
            klmap.xymap.append(list(l[1 : l.rfind(Piece.border)]))
            continue

        if intro and filling_map:
            map_nb = map_nb + 1
            klmap.loadline(map_line_nb, map_name)
            level_by_id[map_nb] = klmap
            klmap = KLMap()
            filling_map = False
            intro = False
            map_name = ''

        if l[0] == '<' and l[-1] == '>':
            intro = True
            map_name = l[1:-1]
            filling_map = True
            map_line_nb = line_nb
            continue

    if filling_map and intro:
        map_nb = map_nb + 1
        klmap.loadline(map_line_nb, map_name)
        level_by_id[map_nb] = klmap

    f.close()
    return level_by_id



'''
Author: Philippe Fremy
License: Gnu GPL (see file LICENSE)
'''
from typing import Union, Dict

from PyQt5.QtWidgets import QDialog, QVBoxLayout, QListView, QWidget
from PyQt5.QtGui import QPixmap
from PyQt5.QtCore import QAbstractListModel, QModelIndex, QVariant, Qt

from .kl_enum import *
from .kl_map import KLMap


class KlMinimapProvider(QAbstractListModel):
    def __init__(self, level_dict: Dict[int, KLMap], mini_maps_dict: Dict[int, QPixmap]) -> None:
        super().__init__()
        assert len(mini_maps_dict) > 0, "minimap board dict is empty!"
        assert len(level_dict) > 0, "level dict is empty!"

        self.level_dict = level_dict
        self.mini_maps_dict = mini_maps_dict

    def rowCount(self, parentIdx: QModelIndex) -> int:  # type: ignore # mypy does not understand the Qt overload of that one
        # -1 because we don't want to display splash screen as a board
        return len(self.level_dict) - 1

    def data(self, index: QModelIndex, role: int = Qt.DisplayRole) -> Union[str, QPixmap, QVariant]:
        if not index.isValid():
            return QVariant()

        row = index.row()
        if role == Qt.DisplayRole:
            return self.level_dict[row + 1].name

        if role == Qt.DecorationRole:
            return self.mini_maps_dict[row + 1]

        return QVariant()


class KLBoardChooser(QDialog):
    def __init__(self, minimapProvider: KlMinimapProvider, parent: QWidget) -> None:
        super().__init__(parent)
        ly = QVBoxLayout(self)
        self.iv = QListView(self)
        self.iv.setModel(minimapProvider)
        ly.addWidget(self.iv, 1)
        self.iv.setViewMode(QListView.IconMode)
        self.iv.setWrapping(True)
        self.iv.setResizeMode(QListView.Adjust)

        self.iv.activated.connect(self.mini_map_selected)
        self.resize(500, 500)

    def mini_map_selected(self, item: QModelIndex) -> None:
        if not item:
            return
        self.done(item.row() + 1)



# type: ignore
'''
Author: Philippe Fremy
License: Gnu GPL (see fname LICENSE)
'''

from src.kl_enum import VERSION

from setuptools import setup

# strip out build status
LONG_DESC = ''.join(open('README.md').readlines()[1:])

print('Packaging klotski version: "%s"' % VERSION)

setup(
    name='Klotski',
    version=VERSION,
    description='A brain-teaser game',
    long_description=LONG_DESC,
    long_description_content_type="text/markdown",
    author='Philippe Fremy',
    author_email='phil.fremy@free.fr',
    url='https://github.com/bluebird75/klotski',
    license='GNU GPL',
    project_urls={
        'Source': 'https://github.com/bluebird75/klotski/',
        'Tracker': 'https://github.com/bluebird75/klotski/issues',
    },
    python_requires='>=3.5',
    install_requires=['PyQt5 >= 5.6', 'sip >= 4.18'],
    packages=['klotski'],
    package_dir={'klotski': 'src'},
    package_data={
        'klotski': ['boards.kts', 'klotski-icon.png', 'klotski-tiles.png', 'README.md'],
    },
    entry_points={
        'gui_scripts': [
            'klotski=klotski.klotski:main',
        ],
    },
    keywords='game',
    classifiers=[
        # see https://pypi.org/classifiers/
        'Development Status :: 5 - Production/Stable',
        # GNU GPL v2 or above
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        # Runs on Windows, unix and MacOs
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        # This version is for python 3 only
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Education',
        'Topic :: Games/Entertainment :: Board Games',
        # Indicate who your project is intended for
        'Intended Audience :: Education',
        'Natural Language :: English',
    ],
)



'''
This is the main file providing the application, the main window and a few
widget.

Author: Philippe Fremy
License: Gnu GPL (see fname LICENSE)
'''

from typing import Optional, Dict, List, Tuple
import sys, pathlib

from PyQt5.QtWidgets import (
    QMainWindow,
    QVBoxLayout,
    QLabel,
    QLCDNumber,
    QApplication,
    QMessageBox,
    QMenu,
    QHBoxLayout,
    QWidget,
    QSizePolicy,
)
from PyQt5.QtGui import QIcon, QPixmap
from PyQt5.QtCore import Qt

from .kl_enum import *
from .kl_map import load_maps, KLMap
from .kl_board import KLBoard
from .kl_board_chooser import KLBoardChooser, KlMinimapProvider


def reverse_move(d: Tuple[int, int]) -> Tuple[int, int]:
    return (-d[0], -d[1])


class Klotski(QMainWindow):
    def __init__(self, maps: Dict[int, KLMap], firstBoard: Optional[int] = None) -> None:
        QMainWindow.__init__(self)
        self.klmap = None  # type: Optional[KLMap]
        self.moves = 0
        self.move_list = []  # type: List[ Tuple[str, Tuple[int, int]]]
        self.move_index = -1
        self.mini_maps_dict = {}  # type: Dict[int, QPixmap]
        self.levels_by_id = maps

        self.init_misc_gui()

        self.board.load_tiles()
        self.board.sig_move.connect(self.move_tile)

        self.generate_mini_maps()

        self.board_chooser = KLBoardChooser(KlMinimapProvider(self.levels_by_id, self.mini_maps_dict), self)
        self.board_chooser.hide()

        # needs self.board
        self.init_menu()

        self.setWindowTitle("Klotski")
        self.title_label.setText("Klotski")
        if firstBoard:
            self.new_level(self.levels_by_id[int(firstBoard)])
            self.move_enabled = True
        else:
            self.new_level(self.levels_by_id[ID_SPLASH_SCREEN])
            self.move_enabled = False

    def init_misc_gui(self) -> None:
        w = QWidget(self)
        self.setCentralWidget(w)
        ly = QVBoxLayout()
        w.setLayout(ly)
        ly.setSpacing(10)

        self.title_label = QLabel("Klotski", w)
        ft = self.title_label.font()
        ft.setPointSize(20)
        self.title_label.setFont(ft)
        self.title_label.setAlignment(Qt.AlignCenter)
        self.title_label.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed))
        ly.addWidget(self.title_label)

        hb = QHBoxLayout()
        self.board = KLBoard(w)
        hb.addWidget(self.board)
        hb.setStretchFactor(self.board, 100)
        ly.addLayout(hb)

        # horizontal bottom hbox [ QLabel | QLcdNumber ]
        hb = QHBoxLayout()
        # hb.setSpacing( 5 )
        # hb.setMargin( 5 )
        ml = QLabel("Moves : ", w)
        ml.setFont(ft)
        ml.setAlignment(Qt.AlignRight)
        hb.addWidget(ml)

        self.move_lcd_nb = QLCDNumber(4, w)
        self.move_lcd_nb.adjustSize()
        self.move_lcd_nb.setFixedSize(self.move_lcd_nb.width(), 30)
        hb.addWidget(self.move_lcd_nb)
        # hb.setFixedHeight( hb.height() )
        ly.addLayout(hb)

        self.adjustSize()

    def generate_mini_maps(self) -> None:
        self.mini_maps_dict = {}
        for i in self.levels_by_id.keys():
            self.mini_maps_dict[i] = self.board.generate_mini_map(self.levels_by_id[i])

    def new_level(self, m: KLMap) -> None:
        '''Display a new map into the main window'''
        if m.name != NAME_SPLASH_SCREEN:
            # display the name of the board
            self.setWindowTitle("Klotski - " + m.name)
            self.title_label.setText(m.name)
        else:
            # display the splash screen
            self.setWindowTitle("Klotski")
            self.title_label.setText('')

        self.klmap = m
        w = max(m.w * TILE_SIZE + 10 + 40, self.title_label.sizeHint().width() + 20)
        h = (
            m.h * TILE_SIZE
            + 10
            + 80
            + self.title_label.sizeHint().height()
            + max(self.move_lcd_nb.sizeHint().height(), self.title_label.sizeHint().height())
        )

        center = self.frameGeometry().center()
        self.resize(w, h)

        r = self.frameGeometry()
        r.moveCenter(center)

        desktop = QApplication.desktop()
        if r.x() + w > desktop.width() or r.x() < 0 or r.y() + h > desktop.height() or r.y() < 0:

            r.moveCenter(desktop.rect().center())

        if r.x() < 0:
            r.setX(0)
        if r.y() < 0:
            r.setY(0)
        if r.x() + w > desktop.width():
            r.setX(0)
            w = desktop.width() - 20 - r.x()
        if r.y() + h > desktop.height():
            r.setY(0)
            h = desktop.height() - 20 - r.y()

        self.move(r.topLeft())
        self.resize(w, h)

        self.reset()

    def choose_board(self) -> None:
        self.board_chooser.exec()
        board_id = self.board_chooser.result()

        if board_id in self.levels_by_id:
            self.new_level(self.levels_by_id[board_id])

    def init_menu(self) -> None:
        file_menu = QMenu('Game', self)
        file_menu.addAction("Boards", self.choose_board, Qt.CTRL + Qt.Key_B)
        file_menu.addAction("Quit", self.close, Qt.CTRL + Qt.Key_X)  # type: ignore
        file_menu.addAction("About Klotski", self.about)

        move_menu = QMenu('Moves', self)
        move_menu.addAction("Reset", self.reset)
        move_menu.addAction("Undo", self.undo, Qt.CTRL + Qt.Key_U)
        move_menu.addAction("Redo", self.redo, Qt.CTRL + Qt.Key_R)

        main_menu = self.menuBar()
        main_menu.addMenu(file_menu)
        main_menu.addMenu(move_menu)

    def set_move_nb(self, m: int) -> None:
        self.move_lcd_nb.display(m)
        self.moves = m

    def move_tile(self, move_info: Tuple[str, Tuple[int, int]]) -> None:
        pid, delta = move_info
        if not self.move_enabled:
            return

        assert self.klmap
        self.klmap.move_piece(pid, delta)
        self.board.move_piece(pid, delta)

        if self.move_index < 0:
            self.move_list = [(pid, (0, 0))]
            self.move_index = 0
            self.set_move_nb(self.moves + 1)

        last_move_pid, last_move = self.move_list[self.move_index]

        if pid != last_move_pid:
            self.move_index = self.move_index + 1
            self.move_list[self.move_index:] = [(pid, delta)]
            self.set_move_nb(self.moves + 1)
        else:
            self.move_list[self.move_index:] = [(pid, (last_move[0] + delta[0], last_move[1] + delta[1]))]

        if self.klmap.is_game_won():
            QMessageBox.information(
                self, "Congratulation", "Congratulations!!!\nYou completed this level in %d moves" % self.moves
            )
            self.move_enabled = False
            self.klmap.reset()

    def reset(self) -> None:
        assert self.klmap
        self.klmap.reset()
        self.board.set_map(self.klmap)
        self.set_move_nb(0)
        self.move_enabled = True
        self.move_list = []
        self.move_index = -1

    def undo(self) -> None:
        if self.move_index < 0:
            return
        if not self.move_enabled:
            return

        pid = self.move_list[self.move_index][0]
        d = reverse_move(self.move_list[self.move_index][1])
        assert self.klmap
        assert self.board
        self.klmap.move_piece(pid, d)
        self.board.move_piece(pid, d)
        self.set_move_nb(self.moves - 1)

        self.move_index = self.move_index - 1

    def redo(self) -> None:
        if self.move_index + 1 >= len(self.move_list):
            return
        if not self.move_enabled:
            return

        self.move_index = self.move_index + 1
        pid, d = self.move_list[self.move_index]
        assert self.klmap
        assert self.board
        self.klmap.move_piece(pid, d)
        self.board.move_piece(pid, d)
        self.set_move_nb(self.moves + 1)

    def about(self) -> None:
        QMessageBox.about(self, 'About Klotski', MSG_ABOUT)

    def print_state(self) -> None:
        print(self.klmap)
        print("Move list : ", self.move_list)
        print("move list index : ", self.move_index)


def main() -> None:
    a = QApplication(sys.argv)
    klotski_icon = QIcon(QPixmap(str(pathlib.Path(__file__).parent / "klotski-icon.ico")))
    QApplication.setWindowIcon(klotski_icon)
    maps = load_maps(str(pathlib.Path(__file__).parent / "boards.kts"))

    firstBoard = None
    if len(sys.argv) > 1:
        firstBoardName = sys.argv[1]
        for m in maps.keys():
            if maps[m].name == firstBoardName:
                firstBoard = m
                break
        else:
            print("No such map: ", firstBoardName)
            print('Map List:')
            for m in maps:
                print(maps[m].name)

    klotski = Klotski(maps, firstBoard)
    klotski.setWindowIcon(klotski_icon)
    klotski.show()
    a.exec()
    del klotski


if __name__ == '__main__':
    main()



'''
This fname provides all the constants and enums other files need to share.

Author: Philippe Fremy
License: Gnu GPL (see fname LICENSE)
'''

from typing import Dict, List

from PyQt5.QtGui import qRgb
from PyQt5.QtCore import Qt

VERSION = '1.3.1'

TILE_SIZE = 32
TRANSP_COLOR = qRgb(0, 0, 0)
ID_SPLASH_SCREEN = 0
NAME_SPLASH_SCREEN = 'Splash'

PIX_SIZE = TILE_SIZE // 2
MINI_TILE_SIZE = 4

TILE_FILE_NAME = "klotski-tiles.png"

MSG_ABOUT = (
    """Klotski %s

by Philippe Fremy <phil.fremy@free.fr>
Graphics by Ben Adler <benadler@bigfoot.de

Klotski is the reborn version of a game existing under Windows 3.1 . This is where all the levels are coming from.

Klotski has been written in 2000 in Python with the graphical toolkit Qt. This was my first big program 
ever written in Python.

For more information about Python, see http://www.python.org
For more information about Qt, see http://www.qt.io.
For more information about PyQt, see https://www.riverbankcomputing.com/software/pyqt/intro
"""
    % VERSION
)


# Piece is used in the map structure
class Piece:
    wall = '#'
    heart = '*'
    pieces = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
    s_wall = '-'
    goal = '.'
    border = '@'
    space = ' '
    none = 'none'


def is_piece(p: str) -> bool:
    '''Analyse the graphic element described by this string and return True if this is a moveable block'''
    for i in p:
        if i not in Piece.pieces:
            return False
    return True


# Tile is use for drawing only
class Tile:
    corner = 0
    hor_edge = 1
    ver_edge = 2
    miss_corner = 3
    inner = 4

    piece = 0
    m_piece = 1
    wall = 2
    s_wall = 3
    goal = 4

    space = 5  # used only for mini tiles color


mini_tile_colors = {
    Tile.piece: Qt.yellow,
    Tile.m_piece: Qt.red,
    Tile.wall: Qt.gray,
    Tile.s_wall: Qt.lightGray,
    Tile.goal: Qt.darkRed,
    Tile.space: Qt.black,
}  # type: Dict[int, Qt.GlobalColor]

pix_list = ["corner", "hor_edge", "ver_edge", "miss_corner", "inner"]  # type: List[str]


tile_list = {
    "piece": Tile.piece,
    "m_piece": Tile.m_piece,
    "wall": Tile.wall,
    "s_wall": Tile.s_wall,
    "goal": Tile.goal,
}  # type: Dict[str, int]


# Translation between map content and tiles
piece_tile = {
    Piece.heart: Tile.m_piece,
    Piece.pieces[0]: Tile.piece,
    Piece.wall: Tile.wall,
    Piece.s_wall: Tile.s_wall,
    Piece.goal: Tile.goal,
    Piece.space: Tile.space,
}  # type: Dict[str, int]
